/***************************************************************************
 *                                                                         *
 *   WEBD.C                                                                *
 *                                                                         *
 *   Copyright (C) 1994 GALACTICOMM, Inc.    All Rights Reserved.          *
 *                                                                         *
 *   World-Wide Web (WWW) server for access to HTML pages.                 *
 *                                                                         *
 *   Ref HTTP/1.0 draft of 3/8/95:                                         *
 *        ftp://ds.internic.net/internet-drafts/                           *
 *        draft-ietf-http-v10-spec-00.txt                                  *
 *                                                                         *
 *   Ref HTML - 2.0 draft of 6/16/95                                       *
 *        ftp://ds.internic.net/internet-drafts/                           *
 *        draft-ietf-html-spec-04.txt                                      *
 *                                                                         *
 *                                              - R. Stein  8/5/94         *
 *   Clickable image maps, statistics logging   - C. Yap    6/95           *
 *   HTTP 1.0, interceptable vectors, forms     - R. Stein  7/95           *
 *                                                                         *
 ***************************************************************************/

#define FILREV "$Revision:   1.0  $"
#define WEBD_VERSION "ICO/2.0"     /* default WEBD_VERSION text variable   */

typedef unsigned char uchar;                     

#include "gcomm.h"
#include "majorbbs.h"
#include "remote.h"
#if defined(UNIX) && defined(SVR4)
#define BAKOPTLEN OPTLEN
#undef OPTLEN
#endif
#include "tcpip.h"
#if defined(UNIX) && defined(SVR4)
#undef OPTLEN
#define OPTLEN BAKOPTLEN
#endif
#include "chandir.h"
#include "inpoly.h"
#include "webd.h"
#include "galwebd.h"
#if defined(UNIX)
#include <fcntl.h>
#include <unistd.h>
#endif
#define byendl(mn) byetcp(mn)                    

STATIC void webincall(int gotchn);
STATIC int webdinp(void);
STATIC void webdsts(void);
STATIC short hdrdone(void);
STATIC short dftwebhdr(void);
STATIC short dfthdrlin(char *hdrlin);
STATIC void dftwebreq(char *uri);
STATIC short dftwebget(char *uri);
STATIC int dftimgmap(char *mapnam,unsigned int xt,unsigned int yt,char **newurl);
STATIC int dftrspbdy(char *buffer,int nbytes);
STATIC int req2mem(void);
STATIC short bdy2fil(void);
STATIC int req2fil(void);
STATIC void rdctfil(char *);
STATIC void webdrst(void);
STATIC char *tvar_now_http(void);
STATIC char *tvar_webd_version(void);

int webdstt;                       /* Web server module state number       */
struct module webdmodule={         /* module interface block               */
     "",                           /*    name used to refer to this module */
     NULL,                         /*    user logon supplemental routine   */
     webdinp,                      /*    input routine if selected         */
     webdsts,                      /*    status-input routine if selected  */
     NULL,                         /*    "injoth" routine for this module  */
     NULL,                         /*    user logoff supplemental routine  */
     NULL,                         /*    hangup (lost carrier) routine     */
     NULL,                         /*    midnight cleanup routine          */
     NULL,                         /*    delete-account routine            */
     NULL                          /*    finish-up (sys shutdown) routine  */
};

struct contype {                   /* content type translate table         */
     char *type;                   /*   type (e.g. 'text/html')            */
     char *ext;                    /*   file extension (e.g. 'htm')        */
     struct contype *next;         /*   link to next entry in table        */
} *contyp=NULL;                    /* head of linked list                  */

STATIC char *months[]={"Jan","Feb","Mar","Apr","May","Jun",
                       "Jul","Aug","Sep","Oct","Nov","Dec"};
STATIC char *dayofwk[]={"Sun","Mon","Tue","Wed","Thu","Fri","Sat"};
STATIC uchar tznames[][4]={
     "PST","PDT","MST","MDT","CST","CDT","EST","EDT","GMT"};
STATIC unsigned int tzmins[]={
      480 , 420 , 420 , 360 , 360 , 300 , 300 , 240 ,  0  };
#define TZNUM (sizeof(tzmins)/sizeof(tzmins[0]))

STATIC void (*oldrst)(void);       /* saved hdlrst() vector                */
STATIC FILE *axslog=NULL;          /* file pointer to access log           */
STATIC FILE *agtlog=NULL;          /* file pointer to agent log            */
STATIC FILE *eralog=NULL;          /* file pointer to error log            */
STATIC char maxwebd;               /* max number of connections to Web svr */
STATIC char webonl;                /* 1=Web server online, 0=offline       */
STATIC char webrej;                /* 1=Reject when offline, 0=ignore      */
STATIC char audweb;                /* audit web retrievals?                */
STATIC char audwebe;               /* audit erroneous web requests?        */
STATIC unsigned int webpatc;       /* patience for activity, 1/16 sec units*/
STATIC FILE *webdmb;               /* GALWEBD.MCV file                     */
char *webroot;                     /* path prefix to all Web HTML pages    */
char *imgroot;                     /* path prefix for image maps           */
char *dftfil;                      /* default file for "GET /" response    */
int webtzone;                      /* minutes behind GMT (240=EDT)         */
char *imgpfix;                     /* prefix for (*hdlimgmap)() i'ceptors  */
                                   /* always begins and ends with "/"      */
uchar *noclick="noclick";          /* pseudo-URI for image maps: ignore clk*/
char *webd_version=WEBD_VERSION;   /* Web server version string (for tvar) */

                                   /*   returns 1=ok, 0=byeweb() called    */
short (*prswebhdr)(void)           /* parse web request header             */
     =dftwebhdr;                   /*   see webdptr->reqsiz and hdrsiz     */

                                   /*   returns 1=handled, 0=not recognized*/
short (*prshdrlin)(                /* parse header line of incoming req    */
     char *hdrlin)=dfthdrlin;      /*   for each header line (not req line)*/

void (*hdlwebreq)(                 /* request interceptor                  */
     char *uri)=dftwebreq;         /*   URI following method in req line   */

                                   /*   ret 1=handled, 0=404 err (no found)*/
short (*hdlwebget)(                /* GET request interceptor              */
     char *uri)=dftwebget;         /*   URI following 'GET' in request line*/

                                   /*   ret 1=mapped, 0=no action, -1=error*/
int (*hdlimgmap)(                  /* Image map mouse-click interceptor    */
     char *mapnam,                 /*   Image map file name (after imgpfix)*/
     unsigned int xt,              /*   pixel X of mouse click             */
     unsigned int yt,              /*   pixel Y of mouse click             */
     char **newurl)=dftimgmap;     /*   where to put pointer to mapped URL */
                                   /*   (when returning 1 only)            */

                                   /*   returns no. bytes sent as rsp body */
int (*hdlrspbdy)(                  /* response body bytes en route client  */
     char *buffer,                 /*   bytes to send in response body     */
     int nbytes)=dftrspbdy;        /*   number of bytes                    */

char *bdyfil="GALWEBD.BDY";        /* temporary file for incom. Entity Body*/
unsigned long bdymax=BDYMAX;       /* maximum size of body (in bdyfil)     */
#define HTTPVLEN (4+1+3+1+2)       /* room for "HTTP/NNN.NN"               */
#define SPRSIZ 120                 /* max assumed size of spr() buffer     */

void EXPORT
init__webd(void)                   /* Initialization                       */
{
     uchar *cp;
     int i;
     static short webinit=0;

     if (webinit) {
          return;
     }
     webinit=1;
     init__tcpip();
     webdmb=opnmsg("galwebd.mcv");
     webonl=ynopt(WEBONL);
     webrej=ynopt(WEBREJ);
     if (webonl || webrej) {
          maxwebd=numopt(MAXWEBD,0,64);
          if (samend(webroot=stgopt(WEBROOT),"\\") || samend(webroot,"/")) {
               webroot[strlen(webroot)-1]='\0';
          }
          if (samend(imgroot=stgopt(IMGROOT),"\\") || samend(imgroot,"/")) {
               imgroot[strlen(imgroot)-1]='\0';
          }
          imgpfix=getmsg(IMGPFIX);
          if (*imgpfix != '/') {
               movmem(imgpfix,imgpfix+1,strlen(imgpfix)+1);
               *imgpfix='/';
          }
          if (!samend(imgpfix,"/")) {
               strcat(imgpfix,"/");
          }
          imgpfix=strdup(imgpfix);
          cp=getmsg(WEBDZONE);
          if (alldgs(cp) || *cp == '-' && alldgs(cp+1)) {
               webtzone=atoi(cp);
          }
          else {
               for (i=0 ; i < TZNUM ; i++) {
                    if (sameas(cp,tznames[i])) {
                         webtzone=tzmins[i];
                         break;
                    }
               }
               if (i == TZNUM) {
                    catastro("Invalid WEBDZONE setting: \"%s\"",cp);
               }
          }
          dftfil=stgopt(DFTFIL);
          audweb=ynopt(AUDWEB);
          audwebe=ynopt(AUDWEBE);
          webpatc=numopt(WEBPATC,0,2047)<<4;
          if (ynopt(AXSFIL)
           && (axslog=fopen(cp=getmsg(AXSFILN),FOPAA)) == NULL) {
               catastro("Cannot write to access log: %s",cp);
          }
          if (ynopt(AGTFIL)
           && (agtlog=fopen(cp=getmsg(AGTFILN),FOPAA)) == NULL) {
               catastro("Cannot write to agent log: %s",cp);
          }
          if (ynopt(ERAFIL)
           && (eralog=fopen(cp=getmsg(ERAFILN),FOPAA)) == NULL) {
               catastro("Cannot write to error log: %s",cp);
          }
          stzcpy(webdmodule.descrp,gmdnam("galwebd.mdf"),MNMSIZ);
          webdstt=register_module(&webdmodule);
          rdctfil(getmsg(CNTFIL));
          regtcpsvr(WEBNAME,(unsigned int)lngopt(WEBPORT,1L,65535L),
                    WEBBACKLOG,webincall);
          oldrst=hdlrst;
          hdlrst=webdrst;
          dclvda(sizeof(struct webdusr)+MINREQBUF);
          register_textvar("NOW_HTTP",tvar_now_http);
          register_textvar("WEBD_VERSION",tvar_webd_version);
     }
     else {
          clsmsg(webdmb);
     }
}

STATIC void
webincall(                         /* begin incoming Web session           */
int gotchn)                        /* 1=chan (curusr) assigned, 0=not avail*/
{                                  /* implicit:  clskt=socket to client    */
     setmbk(webdmb);
     clrprf();
     if (!gotchn) {
          sprintf(prfbuf,xlttxv(stpans(getasc(kilipg || errcod != 1 ?
                                SYSSHUT : SYSFULL)),mxmssz),numcdi("TCP/IP"));
          send(clskt,prfbuf,strlen(prfbuf),0);
     }
     else {
          setmem(webdptr,vdasiz,0);
          usrptr->class=BBSPRV;
          usrptr->state=webdstt;
          usrptr->substt=WEBHDR;
          usrptr->flags|=NOGLOB+NOINJO;
          sprintf(usaptr->userid,"(%s) web browser",inet_ntoa(tcpipinf[usrnum].inaddr));
          btubsz(usrnum,OUTSIZ/2,OUTSIZ/2);
          btutrg(usrnum,OUTSIZ);
          btuech(usrnum,0);
          if (!webonl) {
               byeweb(OFFLINE);
               rejinc(WEBNAME,"Web disabled");
          }
          else if (numonl(webdstt) >= maxwebd) {
               byeweb(WEBFULL,maxwebd);
               rejinc(WEBNAME,spr("%d Web connections",maxwebd));
          }
          else {
               shochl("Incoming Web contact",'W',0x1F);
               sktnfy(TNFRECV,clskt,tcpinc,&tcpipinf[usrnum],usrnum);
               btuinj(usrnum,CYCLE);
               webdptr->sttime=(unsigned int)(hrtval()>>12);
          }
     }
}

STATIC int
webdinp(void)                      /* text line fm client (never happens)  */
{
     return(1);
}

STATIC void
webdsts(void)                      /* status handling                      */
{
     int nbytes;
     int nread;
     int nsent;
     int nroom;
     int more=1;
     unsigned int stnow;

     if (status == RING) {
          rstchn();
          return;
     }
     if (status != CYCLE || (usrptr->flags&BYEBYE)) {
          return;
     }
     setmbk(webdmb);
     stnow=(unsigned int)(hrtval()>>12);
     if (stnow-webdptr->sttime > webpatc && webpatc > 0) {
          byeweb(TIMEOUT,webpatc>>4);
          return;
     }
     switch (usrptr->substt) {
     case WEBHDR:
          if ((nbytes=req2mem()) < 0) {
               more=0;
          }
          else if (nbytes == 0) {
               actdet=0;
          }
          else {
               webdptr->sttime=stnow;
               if (hdrdone()) {
                    more=(*prswebhdr)();
               }
               else if (webdptr->reqcnt == maxreqbuf) {
                    byeweb(HDRBUST,maxreqbuf,maxreqbuf);
                    more=0;
               }
          }
          break;
     case WEBBDY:
          if (webdptr->concnt < webdptr->conlen) {
               nbytes=webdptr->reqfp == NULL ? req2mem() : req2fil();
               if (nbytes < 0) {
                    more=0;
               }
               else if (nbytes == 0) {
                    actdet=0;
               }
               else {
                    webdptr->sttime=stnow;
                    webdptr->concnt+=nbytes;
               }
          }
          else {
               if (webdptr->reqfp != NULL
                && (fflush(webdptr->reqfp) != 0
                 || fseek(webdptr->reqfp,0L,SEEK_SET) != 0)) {
                    byeweb(FRFAIL);
                    more=0;
               }
               else {
                    usrptr->substt=WEBHDL;
               }
          }
          break;
     case WEBHDL:
          btucli(usrnum);
          if (agtlog != NULL && webdptr->userag != NULL) {
               fprintf(agtlog,"%s\n",webdptr->userag);
          }
          (*hdlwebreq)(webdptr->uri);
          more=0;
          break;
     case WEBFIL:
          btucli(usrnum);
          if ((nroom=btuoba(usrnum)) <= 0) {
               actdet=0;
               break;
          }
          if ((nread=fread(vdatmp,1,min(nroom,vdasiz),webdptr->rspfp)) > 0) {
               nsent=(*hdlrspbdy)(vdatmp,nread);
               webdptr->bytcnt+=nsent;
               webdptr->sttime=stnow;
          }
          else if (nroom == OUTSIZ/2-1) {
               endwebd(1);
               byendl(NOMSG);
               more=0;
          }
          else {
               actdet=0;
          }
          break;
     default:
          more=0;
          break;
     }
     if (more) {
          btuinj(usrnum,CYCLE);
     }
}

STATIC short                       /*   1=yes (reqsiz,httpver,hdrsiz set)  */
hdrdone(void)                      /* all of request header received yet?  */
{                                  /*   (inspects httpreqbuf w/o modifying */
     int reqlen;                   /*   req line length (w/o terminators)  */
     char *lfptr;                  /*   pointer to each '\n' of header     */
     int vpos;
     char vbuff[HTTPVLEN+1];
     unsigned int mjr,mnr;
     int beglin;                   /*   index of begin of each header line */

     if ((lfptr=memchr(httpreqbuf,'\n',webdptr->reqcnt)) == NULL) {
          return(0);
     }
     reqlen=(int)(lfptr-httpreqbuf);
     webdptr->reqsiz=reqlen+1;
     if (reqlen > 0 && httpreqbuf[reqlen-1] == '\r') {
          reqlen--;
     }
     webdptr->httpver=0;
     for (vpos=reqlen ; vpos > 0 ; vpos--) {
          if (isspace(httpreqbuf[vpos-1])) {
               if (reqlen-vpos <= HTTPVLEN) {
                    stzcpy(vbuff,httpreqbuf+vpos,reqlen-vpos+1);
                    if (sscanf(vbuff,"HTTP/%u.%u",&mjr,&mnr) == 2) {
                         webdptr->httpver=mjr*100+mnr;
                    }
               }
               break;
          }
     }
     if (webdptr->httpver < 100) {
          webdptr->hdrsiz=webdptr->reqsiz;
          return(1);
     }
     while ((beglin=(int)(lfptr-httpreqbuf)+1) < webdptr->reqcnt) {
          if (httpreqbuf[beglin] == '\r') {       /* CR of hdr terminator? */
               if (++beglin >= webdptr->reqcnt) {
                    return(0);
               }
          }
          if (httpreqbuf[beglin] == '\n') {       /* LF of hdr terminator? */
               webdptr->hdrsiz=beglin+1;
               return(1);
          }
          if ((lfptr=memchr(httpreqbuf+beglin,'\n',
                            webdptr->reqcnt-beglin)) == NULL) {
               return(0);
          }
     }
     return(0);
}

STATIC short                       /*   returns 1=ok, 0=byeweb() called    */
dftwebhdr(void)                    /* default (*prswebhdr)() vector        */
                                   /* Parse web request header and decide  */
                                   /* if a body is to be received (conlen).*/
                                   /* Expects reqsiz & hdrsiz already set. */
                                   /* If ret true, computes & terminates   */
                                   /* these fields:  method,uri,vertxt.    */
                                   /* If ret true, also sets new substt.   */
{                                  /* First vector called for each new req.*/
     char *hdrlin;
     char *remhdr;

     setmbk(webdmb);
     usrptr->substt=WEBHDL;
     httpreqbuf[webdptr->reqsiz-1]='\0';     /* terminate request line     */
     if ((webdptr->method=strtok(httpreqbuf," \t\r\n")) == NULL
      || (webdptr->uri=strtok(NULL," \t\r\n")) == NULL) {
          byeweb(REQFMT);
          return(0);
     }
     if ((webdptr->vertxt=strtok(NULL," \t\r\n")) == NULL) {
          webdptr->vertxt="";
     }
     if (webdptr->hdrsiz == webdptr->reqsiz) {
          return(1);
     }
     httpreqbuf[webdptr->hdrsiz-1]='\0';     /* terminate header           */
     for (hdrlin=strtok(httpreqbuf+webdptr->reqsiz,"\r\n")
        ; hdrlin != NULL
        ; hdrlin=strtok(remhdr,"\r\n")) {
          remhdr=strtok(NULL,"");
          (*prshdrlin)(hdrlin);
     }
     setmbk(webdmb);
     if (webdptr->conlen > 0) {
          if (webdptr->conlen > bdymax) {
               byeweb(BDYBUST,ul2as(bdymax),ul2as(bdymax));
               return(0);
          }
          if (webdptr->conlen > maxreqbuf-webdptr->hdrsiz) {
               if (!bdy2fil()) {
                    return(0);
               }
          }
          usrptr->substt=WEBBDY;
     }
     return(1);
}

STATIC short                       /*   returns 1=handled, 0=not recognized*/
dfthdrlin(                         /* default (*prshdrlin)() vector        */
char *hdrlin)                      /*   line from the incoming req header  */
{
     char *httpdat;

     setmbk(webdmb);
     if (isspace(hdrlin[0])) {
     }
     else if (sscanf(hdrlin,"Content-Length: %ul",&webdptr->conlen) == 1) {
     }
     else if (hdrstg(hdrlin,"Content-Type:",&webdptr->contyp)) {
     }
     else if (hdrstg(hdrlin,"If-Modified-Since:",&httpdat)) {
          if (dechttpdate(httpdat,&webdptr->imsdat,&webdptr->imstim)) {
               webdptr->flags|=WFLIMS;
          }
     }
     else if (hdrstg(hdrlin,"Referer:",&webdptr->refer)) {
     }
     else if (hdrstg(hdrlin,"User-Agent:",&webdptr->userag)) {
     }
     else {
          return(0);
     }
     return(1);
}

STATIC void
dftwebreq(                         /* default (*hdlwebreq)() vector        */
char *uri)                         /*   URI from request line              */
{
     unsigned int date;

     setmbk(webdmb);
     if (sameas(webdptr->method,"GET")) {
          if (!(*hdlwebget)(uri)) {
               setmbk(webdmb);
               byeweb(NOTFND,uri,uri);
               if (audwebe) {
                    shocst("WEB SERVER PAGE DOES NOT EXIST",
                           "page \"%0.60s\" requested by %s",
                           uri,inet_ntoa(tcpipinf[usrnum].inaddr));
               }
               if (eralog != NULL) {
                    date=today();
                    fprintf(eralog,"[%s %s %d %s %d] "
                            "httpd: access to %s failed for %s reason: "
                            "file does not exist%s\n",
                            dayofwk[daytoday()],months[ddmon(date)-1],
                            ddday(date),nctime(now()),ddyear(date),
                            uri,inet_ntoa(tcpipinf[usrnum].inaddr),
                            webdptr->refer == NULL ? ""
                            : spr(" from %-0.100s",webdptr->refer));
               }
          }
          setmbk(webdmb);
     }
     else {
          byeweb(BADMETH,webdptr->method,webdptr->method);
     }
}

STATIC short                       /*   ret 1=handled, 0=404 err (no found)*/
dftwebget(                         /* default (*hdlwebget)() routine       */
char *uri)                         /*   resource identifier                */
{
     return(mapwebget(uri) || filwebget(uri));
}

short                              /*   ret 1=handled, 0=not recognized    */
mapwebget(                         /* NCSA-style image map file interpreter*/
char *uri)                         /* URI from client                      */
{
     char *qpos;
     unsigned int xt,yt;
     char *mapuri;
     char *newurl;
     int imgrc;

     setmbk(webdmb);
     if (!sameto((uchar *)imgpfix,(uchar *)uri)) {
          return(0);
     }
     mapuri=uri+strlen(imgpfix)-1;      /* mapuri should start with a "/"  */
     if ((qpos=strrchr(mapuri,'?')) == NULL
      || sscanf(qpos+1,"%u,%u",&xt,&yt) != 2) {
          byeweb(BADMAP,imgpfix);
          return(1);
     }
     *qpos='\0';                        /* mapuri is now just map file name*/
     imgrc=(*hdlimgmap)(mapuri,xt,yt,&newurl);
     setmbk(webdmb);
     switch (imgrc) {
     case -1:
          break;
     case 0:
          byeweb(MAPNOCLK);
          break;
     case 1:
          if (strchr(newurl,':') == NULL) {
               if (!filwebget(newurl)) {
                    byeweb(BADMAPRF,newurl,mapuri,newurl,mapuri);
               }
          }
          else {
               byeweb(MAPREDIR,newurl,newurl,newurl);
          }
          break;
     }
     return(1);
}

short                              /*   ret 1=handled, 0=404 err (no found)*/
filwebget(                         /* file-retrieval form of GET request   */
char *uri)                         /*   URI (file path rel to WEBROOT{})   */
{                                  /*   (corrupts vdatmp)                  */
     struct fndblk fb;
     char *pth,*cp;

     setmbk(webdmb);
     if (sameas(uri,"/")) {
          uri=dftfil;
     }
     if (badfspec(uri)) {
          return(1);
     }
     pth=webpth(uri);
     if (!fnd1st(&fb,pth,0)) {
          return(0);
     }
     if (webdptr->flags&WFLIMS
      && (fb.date < webdptr->imsdat
       || (fb.date == webdptr->imsdat && fb.time < webdptr->imstim))) {
          byeweb(NOTMOD);
          return(1);
     }
     if ((webdptr->rspfp=fopen(pth,FOPRB)) == NULL) {
          return(0);
     }
     if (webdptr->httpver >= 100) {
          if ((cp=getctyp(fnext(fb.name))) != NULL) {
               prf(xlttxv(stpans(getasc(FILEGET2)),mxmssz),
                   ul2as(fb.size),enchttpdate(fb.date,fb.time),
                   cp);
          }
          else {
               prf(xlttxv(stpans(getasc(FILEGET)),mxmssz),
                   ul2as(fb.size),enchttpdate(fb.date,fb.time));
          }
          powprf();
     }
     usrptr->substt=WEBFIL;
     btuinj(usrnum,CYCLE);
     shochl("Web page retrieval",'W',0x1F);
     return(1);
}

STATIC int                         /*   ret 1=mapped, 0=no action, -1=error*/
dftimgmap(                         /* default (*hdlimgmap)() handler       */
char *mapnam,                      /*   file name for image map            */
unsigned int xt,                   /*   pixel X of mouse click             */
unsigned int yt,                   /*   pixel Y of mouse click             */
char **newurl)                     /*   where to put pointer to mapped URL */
                                   /*   (useful when returning 1 only)     */
{                                  /*   (corrupts vdatmp)                  */
     char *type;
     char *url;
     static unsigned int poly[MAXVERT][2];
     #define x1 (poly[0][0])
     #define y1 (poly[0][1])
     #define x2 (poly[1][0])
     #define y2 (poly[1][1])
     int npoints;
     int n;
     unsigned long mindist;
     unsigned long pntdist;
     FILE *mapfp;
     char *wordn;
     int bingo;                    /* definite hit                         */
     int pbingo;                   /* potential hit, see if better one come*/
     char *pth;

     setmbk(webdmb);
     if (badfspec(mapnam)) {
          return(-1);
     }
     pth=imgpth(mapnam);
     if ((mapfp=fopen(pth,FOPRA)) == NULL) {
          byeweb(MAPNOT,pth,pth);
          return(-1);
     }
     pbingo=0;
     while (fgets(vdatmp,vdasiz-URISIZ,mapfp) != NULL) {
          if ((n=strlen(vdatmp)) > 0 && vdatmp[n-1] == '\n') {
               vdatmp[n-1]='\0';
          }
          if (vdatmp[0] != '#'
           && vdatmp[0] != '\0'
           && (type=strtok(vdatmp," \t")) != NULL
           && (url=strtok(NULL," \t")) != NULL) {
               for (npoints=0 ; npoints < MAXVERT ; npoints++) {
                    if ((wordn=strtok(NULL," \t")) == NULL
                     || sscanf(wordn,"%u,%u",&poly[npoints][0],
                                             &poly[npoints][1]) != 2) {
                         break;
                    }
               }
               bingo=0;
               if (sameas((uchar *)type,(uchar *)"default") && !pbingo) {
                    *newurl=stzcpy(vdatmp+vdasiz-URISIZ,url,URISIZ);
                    mindist=(unsigned long)(-1L);
                    pbingo=1;
               }
               else if (sameto((uchar *)"rect",(uchar *)type)
                     && npoints == 2) {
                    bingo=(x1 <= xt) == (xt <= x2)
                       && (y1 <= yt) == (yt <= y2);
               }
               else if (sameto((uchar *)"circ",(uchar *)type)
                     && npoints == 2) {
                    bingo=udif2(x1,xt)+udif2(y1,yt)
                        < udif2(x1,x2)+udif2(y1,y2);
               }
               else if (sameto((uchar *)"poly",(uchar *)type)
                     && npoints >= 3) {
                    bingo=inpoly(poly,npoints,xt,yt);
               }
               else if (sameto((uchar *)"point",(uchar *)type)
                     && npoints == 1) {
                    pntdist=udif2(xt,x1)+udif2(yt,y1);
                    if (!pbingo) {
                         mindist=pntdist;
                         *newurl=stzcpy(vdatmp+vdasiz-URISIZ,url,URISIZ);
                         pbingo=1;
                    }
                    else if (pntdist < mindist) {
                         mindist=pntdist;
                         *newurl=stzcpy(vdatmp+vdasiz-URISIZ,url,URISIZ);
                    }
               }
               if (bingo) {
                    *newurl=url;
                    fclose(mapfp);
                    return(!sameas((uchar *)(*newurl),noclick));
               }
          }
     }
     fclose(mapfp);
     return(pbingo && !sameas((uchar *)(*newurl),noclick));
     #undef x1
     #undef y1
     #undef x2
     #undef y2
}

STATIC int                         /*   returns no. bytes sent in response */
dftrspbdy(                         /* default (*hdlrspbdy)() vector        */
char *buffer,                      /*   bytes to send in response body     */
int nbytes)                        /*   number of bytes                    */
{
     int nroom,nsent;

     setmbk(webdmb);
     nroom=btuoba(usrnum);
     nsent=min(nroom,nbytes);      /* recompute in case intercepted        */
     if (nsent > 0) {
          btuxct(usrnum,nsent,buffer);
     }
     return(nsent);
}

STATIC int                         /*   0=no data, -1=byeweb(), else=nbytes*/
req2mem(void)                      /* receive request bytes to httpreqbuf  */
{
     int ibw,nroom,ica;

     if ((ibw=btuibw(usrnum)) <= 0) {
          return(0);
     }
     nroom=maxreqbuf-webdptr->reqcnt;
     ASSERT(nroom > 0);
     if ((ica=btuica(usrnum,httpreqbuf+webdptr->reqcnt,min(ibw,nroom))) < 0) {
          byeweb(ICAGOOF);
          return(-1);
     }
     webdptr->reqcnt+=ica;
     ASSERT(webdptr->reqcnt <= maxreqbuf);
     return(ica);
}

STATIC short                       /*   returns 1=success, 0=byeweb called */
bdy2fil(void)                      /* write request body (so far) to file  */
{                                  /*   leaves file open for read & write  */
     unsigned int nwrite;

     chdmak();
     ASSERT(webdptr->reqfp == NULL);
     unlink(chdpath(bdyfil));
     if ((webdptr->reqfp=fopen(chdpath(bdyfil),FOPRWB)) == NULL) {
          byeweb(FOPFAIL,bdyfil,bdyfil);
          return(0);
     }
     if (webdptr->reqcnt > webdptr->hdrsiz) {
          webdptr->concnt=nwrite=webdptr->reqcnt-webdptr->hdrsiz;
          if (fwrite(httpreqbuf+
                     webdptr->hdrsiz,1,nwrite,webdptr->reqfp) != nwrite) {
               byeweb(FWFAIL);
               return(0);
          }
     }
     return(1);
}

STATIC int                         /*   0=no data, -1=byeweb(), else=nbytes*/
req2fil(void)                      /* receive request bytes to body file   */
{                                  /*   (corrupts vdatmp)                  */
     int ica;

     ASSERT(webdptr->reqfp != NULL);
     if ((ica=btuica(usrnum,vdatmp,vdasiz)) < 0) {
          byeweb(ICAGOOF);
          return(-1);
     }
     if (ica == 0) {
          return(0);
     }
     if (fwrite(vdatmp,1,ica,webdptr->reqfp) != ica) {
          byeweb(FWFAIL);
          return(-1);
     }
     return(ica);
}

/*--- Web Server utilities ---*/

short                              /*   ret 1=bad (byeweb() called), 0=ok  */
badfspec(                          /* Is URI-based filespec dangerous?     */
char *uri)                         /*   URI or filespec based on a URI     */
{
     if (strstr(uri,"..") != NULL
      || strchr(uri,':') != NULL
      || rsvnam(uri)) {
          byeweb(BADFSPEC,uri);
          return(1);
     }
     return(0);
}

short                              /*   returns 1=match, 0=not             */
hdrstg(                            /* parse a named token                  */
char *source,                      /*   line, eg 'Content-Type: text/html' */
char *token,                       /*   token name, eg 'Content-Type:'     */
char **var)                        /*   where to store pointer to value    */
{
     char *retval;

     if (sameto(token,source)) {
          retval=source+strlen(token);
          while (isspace(*retval)) {
               retval++;
          }
          *var=retval;
          return(1);
     }
     return(0);
}

unsigned long
udif2(                             /* square of difference of two u-ints   */
unsigned int a,
unsigned int b)
{
     unsigned long dif;

     dif=(a < b) ? b-a : a-b;
     return(dif*dif);
}

char *
webpth(                            /* format path name of web page         */
char *pagnam)                      /* page name (can start with "/" or "\")*/
{                                  /* returns "" if combination too long   */
     if (strlen(webroot)+1+strlen(pagnam)+1 > SPRSIZ) {
          return("");
     }
     return(spr(pagnam[0] == '\\'
             || pagnam[0] == '/' ? "%s%s" : "%s/%s",webroot,pagnam));
}

char *
imgpth(                            /* format path name of web page         */
char *pagnam)                      /* page name (can start with "/" or "\")*/
{                                  /* returns "" if combination too long   */
     if (strlen(imgroot)+1+strlen(pagnam)+1 > SPRSIZ) {
          return("");
     }
     return(spr(pagnam[0] == '\\'
             || pagnam[0] == '/' ? "%s%s" : "%s/%s",imgroot,pagnam));
}

STATIC void
rdctfil(                           /* Reads in the type and extensions     */
                                   /*   for the Content-Type field         */
char *filnam)                      /*   name of file to read from          */
{
     struct contype *curr;
     FILE *cntfil;
     char *type,*ext,*rbuf;
     unsigned int linnum;

     if ((cntfil=fopen(filnam,FOPRA)) == NULL) {
          catastro("webd: Cannot open Content-Type file %s",filnam);
     }
     rbuf=(char *)alcmem(512);
     linnum=0;
     while (fgets(rbuf,512,cntfil) != NULL) {
          linnum++;
          if (rbuf[0] != '#' && !isspace(rbuf[0])) {
               if ((type=strtok(rbuf," \t\r\n")) == NULL
                || (ext=strtok(NULL," ,\t\r\n")) == NULL) {
                    catastro("Web Content-Type file %s line %u:  bad syntax",
                             filnam,linnum);
               }
               if (getctyp(ext) != NULL) {
                    catastro("Web Content-Type file %s line %u:\n"
                             "extension \"%s\" defined twice",
                             filnam,linnum,ext);
               }
               curr=(struct contype *)alcmem(sizeof(struct contype));
               curr->type=(char *)alcmem(strlen(type)+1);
               curr->ext=(char *)alcmem(strlen(ext)+1);
               strcpy(curr->type,type);
               strcpy(curr->ext,ext);
               curr->next=contyp;
               contyp=curr;
          }
     }
     free(rbuf);
     fclose(cntfil);
}

char *                             /*   ret Content-Type (NULL=not found)  */
getctyp(                           /* get Content-Type from global list    */
char *ext)                         /*   extension to search for in list    */
{
     struct contype *curr;

     if (ext == NULL) {
          return(ext);
     }
     for (curr=contyp ; curr ; curr=curr->next) {
          if (sameas(curr->ext,ext)) {
               return(curr->type);
          }
     }
     return((char *)NULL);
}

/*--- Web Server end of session ---*/

void
byeweb(msgnum,p1,p2,p3)            /* say goodbye to a Web user w/msg      */
int msgnum;                        /* same parameters as prfmsg()          */
long p1,p2,p3;                     /* usrnum is an implicit input          */
{                                  /* (do not use msgnum=NOMSG or PAMSG)   */
     ASSERTM(btuoba(usrnum) == OUTSIZ/2-1,
             spr("msgnum=%d,btuoba=%d",msgnum,btuoba(usrnum)));
     endwebd(0);
     btulok(usrnum,1);
     btucli(usrnum);
     btuclo(usrnum);
     btuoes(usrnum,1);
     prfmsg(msgnum,p1,p2,p3);
     stpans(prfbuf);
     outprf(usrnum);
     clrprf();
     user[usrnum].flags|=BYEBYE;
}

void
endwebd(                           /* terminate the WWW session            */
int complt)                        /* 1=finished, 0=aborted                */
{
     short date;

     btubsz(usrnum,INPSIZ,OUTSIZ);
     if (webdptr->reqfp != NULL) {
          fclose(webdptr->reqfp);
          unlink(chdpath(bdyfil));
          webdptr->reqfp=NULL;
     }
     if (webdptr->rspfp != NULL) {
          fclose(webdptr->rspfp);
          webdptr->rspfp=NULL;
          if (audweb && !(webdptr->flags&WFLERR)) {
               shocst(complt ? "WEB SERVER PAGE RETRIEVED"
                             : "WEB SERVER RETRIEVAL ABORTED",
                      "page \"%s\", %ld bytes, sent to %s",
                      webdptr->uri,webdptr->bytcnt,
                      inet_ntoa(tcpipinf[usrnum].inaddr));
          }
     }
     if (axslog != NULL && complt && !(webdptr->flags&WFLERR)) {
          date=today();
          fprintf(axslog,"%s - - [%d/%s/%d:%s -0000] \"%s %s %s\" 200 %d\n",
               inet_ntoa(tcpipinf[usrnum].inaddr),ddday(date),
               months[ddmon(date)-1],ddyear(date),nctime(now()),
               webdptr->method,webdptr->uri,webdptr->vertxt,webdptr->bytcnt);
     }
}

STATIC void
webdrst(void)                      /* Web server reset interception        */
{
     if (usrptr->class == BBSPRV && usrptr->state == webdstt) {
          endwebd(0);
     }
     (*oldrst)();
}

/*--- Web Server text variables ---*/

STATIC char *
tvar_now_http(void)                /* text variable:  RFC1123 fmt date&time*/
{
     return(enchttpdate(today(),now()));
}

STATIC char *
tvar_webd_version(void)            /* text variable:  Web server version   */
{
     return(webd_version);
}

/*--- The following routine, tcpinc(), will be moved to TCPIP.C in ICO  ---*/
/*--- version 2.0.  It's included here for binary compatibility w/1.0.  ---*/

void
tcpinc(                            /* read socket data into GCDI channel   */
struct tcpipinf *tip)              /* user's TCP/IP channel info           */
                                   /* application:  sktnfy(,,tcpinc,tip,)  */
{
     struct datstm *dsp;
     unsigned nroom;
     int nactual;
     int try;

     dsp=tip->cdidst;
     nroom=dsp->snkwin(dsp);
     ASSERT(tip->socket == nfyskt);
     for (try=0 ; try < 10 ; try++) {
          if ((nroom=dsp->snkwin(dsp)) == 0) {
               if (try == 0) {
                    actdet=0;
               }
               break;
          }
          if ((nactual=recv(tip->socket,dsp->snkloc,nroom,0)) < 0) {
               switch (tcpip_errno) {
               case ECONNRESET:         /* benign disconnect               */
                    sktcnc(TNFRECV,tip->socket);
                    bbsdsc(usrnum);
                    break;
               case EWOULDBLOCK:        /* no more polling necessary       */
                    break;
               default:                 /* error condition                 */
                    svrfail(usrnum,"RECEIVE",tip->server->port);
                    break;
               }
               break;
          }
          if (nactual == 0) {
               if (try == 0) {
                    bbsdsc(usrnum);
               }
               break;
          }
          dsp->didmov(dsp,nactual);
          if (tip->outsnk.bufcnt > 0) {
               tcpsnd(tip);
          }
          if (nactual < nroom) {   /* recv() didn't exhaust sink's window, */
               break;              /* no point in trying to recv() again   */
          }
     }
}

/*--- The following code may one day be moved to a more general purpose ---*/
/*--- place, such as DOSFACE.C or DATUTILS.C or DSKUTL.C.               ---*/

#define DAYSECS (60UL*60UL*24UL)   /* seconds in a day                     */
#define DECADESECS (DAYSECS*(365UL*10UL+2UL))
                                   /* seconds from 1/1/70 to 1/1/80        */

void
datu2d(                            /* convert Unix date/time to DOS format */
unsigned long uxtime,              /*   Unix seconds since 1/1/70          */
unsigned short *dat,               /*   pointer to DOS packed date output  */
unsigned short *tim)               /*   pointer to DOS packed time output  */
{
     unsigned short day,hour,minsec;

     if (uxtime >= DECADESECS) {
          uxtime-=DECADESECS;      /* conv to secs since 00:00 1/1/80      */
     }
     else {
          uxtime=0UL;
     }
     day=(unsigned short)(uxtime/DAYSECS);            /* days since 1/1/80 */
     hour=(unsigned short)((uxtime/3600UL)%24UL);     /* hour since midnite*/
     minsec=(unsigned short)(uxtime%3600UL);          /* seconds since hour*/
     *tim=dttime(hour,minsec/60,minsec%60);
     *dat=datofc(day);
}

void
datd2u(                            /* convert DOS date/time to Unix format */
unsigned short dat,                /*   DOS packed date                    */
unsigned short tim,                /*   DOS packed time                    */
unsigned long *uxtime)             /*   ptr, Unix secs since 1/1/70 output */
{
     *uxtime=DECADESECS+           /* 00:00 1/1/70 to 00:00 1/1/80         */
             cofdat(dat)*DAYSECS+  /* 00:00 1/1/80 to 00:00 fileday        */
             dthour(tim)*3600UL+   /*  \                                   */
             dtmin(tim)*60UL+      /*   > compute filetime since           */
             dtsec(tim);           /*  /  00:00 on fileday                 */
}

char *                             /*   returns pointer to formatted date  */
enchttpdate(                       /* Encode an HTTP date                  */
unsigned short dat,                /*   DOS packed date format             */
unsigned short tim)                /*   DOS packed time format             */
{
     static char buff[40];
     unsigned long uxnow;

     datd2u(today(),now(),&uxnow);
     datu2d(uxnow+webtzone*60,&dat,&tim);
     sprintf(buff,"%-3.3s, %02u %s %u %02u:%02u:%02u GMT",
                  dayofwk[(cofdat(dat)+2)%7],
                  ddday(dat),moname[ddmon(dat)],ddyear(dat),
                  dthour(tim),dtmin(tim),dtsec(tim));
     buff[9]=tolower(buff[9]);
     buff[10]=tolower(buff[10]);
     return(buff);
}

int                                /*   returns 1=ok, 0=bad format         */
dechttpdate(                       /* Decode an HTTP date                  */
char *dnt,                         /*   RFC 1123, RFC 850, or ANSI C format*/
unsigned short *dat,               /*   DOS packed date format             */
unsigned short *tim)               /*   DOS packed time format             */
                                   /*   CAUTION: destroys dnt date & time  */
                                   /*   See HTTP 1.0 section 3.3.1, 3/8/95 */
                                   /*   1123: WWW, DD MMM YYYY HH:MM:SS GMT*/
                                   /*   850: WWWWWW, DD-MMM-YY HH:MM:SS GMT*/
                                   /*   ANSI: WWW MMM DD HH:MM:SS YYYY     */
{                                  /*   Good for dates from 1980 to 2069   */
     unsigned short num[6];        /* decoded numbers                      */
     static char dlims[]=" -:";    /* delimiters between numbers & words   */
     int numstg;                   /* qty of non-numbers (exc day of week) */
     int idxstg;                   /* index, 0 to 5, of last non-number    */
     uchar *stg;                   /* last non-number                      */
     int hri;                      /* index, 0-5, of hour (min&sec follow) */
     unsigned short mon,day,year;  /* values extracted from numbers & words*/
     uchar *cp;
     int i;
     unsigned long uxtim;

     cp=strtok(dnt,dlims);
     if (cp == NULL) {
          return(0);
     }
     numstg=0;
     for (i=0 ; i < 6 ; i++) {
          if ((cp=strtok(NULL,dlims)) == NULL) {
               return(0);
          }
          if (alldgs(cp)) {
               num[i]=(unsigned short)atoi(cp);
          }
          else {
               numstg++;
               idxstg=i;
               stg=cp;
          }
     }
     if (numstg != 1 || idxstg >= 2) {
          return(0);
     }
     for (i=1 ; i <= 12 ; i++) {
          if (sameto((uchar *)moname[i],stg)) {
               mon=i;
               break;
          }
     }
     if (i > 12) {
          return(0);
     }
     if (idxstg == 1) {            /* RFC 1123 or 850 format date & time   */
          year=num[2];
          day=num[0];
          hri=3;
     }
     else {                        /* ANSI C's asctime() format            */
          year=num[5];
          day=num[1];
          hri=2;
     }
     if (year < 70) {                             /* 00-69 => 2000-2069    */
          year+=2000;                             /* 70-79 => 1980         */
     }                                            /* 80-99 => 1980-1999    */
     if (year < 100) {                            /* 1970-1979 => 1980     */
          year+=1900;                             /* 1980-2069 => 1980-2069*/
     }
     if (year < 1980) {
          year=1980;
     }
     *dat=dddate(mon,day,year);
     *tim=dttime(num[hri],num[hri+1],num[hri+2]);
     datd2u(*dat,*tim,&uxtim);
     datu2d(uxtim-webtzone*60,dat,tim);
     return(1);
}

